"""Module to support Super-B v2 BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Super-B v2 BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Super-B", "default_model": "Epsilon v2"}
    _NOTIFY_CHAR: Final[str] = "e0fef452-9d2b-4005-a1e3-69fe1102b436"
    _FRAME_LEN: Final[int] = 24
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("current", 6, 4, True, lambda x: x / 1000, 0x2),
        BMSDp("voltage", 10, 2, False, lambda x: x / 1000, 0x2),
        BMSDp("battery_level", 1, 1, False, idx=0x0),
        BMSDp("battery_health", 19, 1, False, idx=0x0),
        BMSDp("runtime", 20, 4, False, idx=0x0),
        BMSDp("cycles", 18, 1, False, idx=0x0),
        # BMSDp("problem", ?, 1, False, lambda x: not (x & 1), 0x0),
        # BMSDp("balancer", ?, 1, False, lambda x: bool(x & 0x80), 0x0),
    )
    _CMDS: Final[set[int]] = {field.idx for field in _FIELDS}

    def __init__(
        self,
        ble_device: BLEDevice,
        keep_alive: bool = True,
        secret: str = "",
        logger_name: str = "",
    ) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive, secret, logger_name)
        self._msg: dict[int, bytes] = {}

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            MatcherPattern(
                local_name="Epsilon*",
                manufacturer_id=0x50BE,
                manufacturer_data_start=list(b"Epsilon V2"),
                connectable=True,
            )
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return ("cf9ccdf7-eee9-43ce-87a5-82b54af5324e",)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "cf9ccdfa-eee9-43ce-87a5-82b54af5324e"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "cf9ccdfa-eee9-43ce-87a5-82b54af5324e"

    @staticmethod
    def _raw_values() -> frozenset[BMSValue]:
        return frozenset({"runtime"})  # never calculate runtime

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        await super()._init_connection(char_notify)
        # subscribe to second notify characteristic
        await self._client.start_notify(
            BMS._NOTIFY_CHAR, getattr(self, "_notification_handler")
        )

    def _notification_handler(
        self, sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data from %s: %s", str(sender)[:8], data)

        if len(data) != BMS._FRAME_LEN:
            self._log.debug("incorrect frame length")
            return

        self._msg[data[0]] = bytes(data)
        if BMS._CMDS.issubset(self._msg.keys()):
            self._msg_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_msg(b"\x21\x54\x00")
        result: BMSSample = self._decode_data(BMS._FIELDS, self._msg)

        # remove runtime if not discharging
        if result.get("runtime", 0) == 0xFFFFFFFF:
            result.pop("runtime", None)
        self._msg.clear()
        return result



"""Test the E&J technology BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ej_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 20


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockEJBleakClient(MockBleakClient):
    """Emulate a E&J technology BMS BleakClient."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("6e400002-b5a3-f393-e0a9-e50e24dcca9e"):
            return bytearray()
        cmd: int = int(bytearray(data)[3:5], 16)
        if cmd == 0x02:
            return bytearray(
                b":0082310080000101C00000880F540F3C0F510FD70F310F2C0F340F3A0FED0FED0000000000000000"
                b"000000000000000248424242F0000000000000000001AB~"
            )  # TODO: put numbers
        if cmd == 0x10:
            return bytearray(b":009031001E00000002000A000AD8~")  # TODO: put numbers
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        self._notify_callback("MockEJBleakClient", bytearray(b"AT\r\n"))
        self._notify_callback("MockEJBleakClient", bytearray(b"AT\r\nillegal"))
        for notify_data in [
            self._response(char_specifier, data)[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._response(char_specifier, data)), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockEJBleakClient", notify_data)


class MockEJsfBleakClient(MockEJBleakClient):
    """Emulate a E&J technology BMS BleakClient with single frame protocol."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("6e400002-b5a3-f393-e0a9-e50e24dcca9e"):
            return bytearray()

        if int(bytearray(data)[3:5], 16) == 0x02:
            return bytearray(
                b":008231008C000000000000000CBF0CC00CEA0CD50000000000000000000000000000000000000000"
                b"00000000008C000041282828F000000000000100004B044C05DC05DCB2~"
            )
        return bytearray()

    @staticmethod
    def values() -> BMSSample:
        """Return correct data sample values for single frame protocol sample."""
        return {
            "voltage": 13.118,
            "current": 1.4,
            "battery_level": 75,
            "cycles": 1,
            "cycle_charge": 110.0,
            "cell_count": 4,
            "cell_voltages": [3.263, 3.264, 3.306, 3.285],
            "delta_voltage": 0.043,
            "temperature": 25,
            "temp_values": [25],
            "cycle_capacity": 1442.98,
            "power": 18.365,
            "battery_charging": True,
            "problem": False,
            "problem_code": 0,
            "balancer": 0,
            "chrg_mosfet": True,
            "dischrg_mosfet": True,
            "heater": False,
            "design_capacity": 150,
        }


class MockEJsfnoCRCBleakClient(MockEJsfBleakClient):
    """Emulate a E&J technology BMS BleakClient with single frame protocol and uncalculated CRC."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        ret: bytearray = MockEJsfBleakClient._response(self, char_specifier, data)
        ret[-3:-1] = b"00"  # patch to wrong CRC
        return ret


class MockEJinvalidBleakClient(MockEJBleakClient):
    """Emulate a E&J technology BMS BleakClient without sending second frame."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("6e400002-b5a3-f393-e0a9-e50e24dcca9e"):
            return bytearray()

        return bytearray(
            b":0082310080000101C00000880F540F3C0F510FD70F310F2C0F340F3A0FED0FED0000000000000000"
            b"000000000000000248424242F0000000000000000001AB~"
        )  # TODO: put numbers


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test E&J technology BMS data update."""

    patch_bleak_client(MockEJBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == {
        "voltage": 39.517,
        "current": -0.02,
        "battery_level": 1,
        "cycles": 0,
        "cycle_charge": 0.2,
        "cell_count": 10,
        "cell_voltages": [
            3.924,
            3.900,
            3.921,
            4.055,
            3.889,
            3.884,
            3.892,
            3.898,
            4.077,
            4.077,
        ],
        "delta_voltage": 0.193,
        "temperature": 32,
        "temp_values": [32],
        "cycle_capacity": 7.903,
        "power": -0.79,
        "runtime": 36000,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "heater": False,
    }

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockEJBleakClient, BMS)


async def test_update_single_frame(
    patch_bleak_client, keep_alive_fixture: bool
) -> None:
    """Test E&J technology BMS data update."""

    patch_bleak_client(MockEJsfBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == MockEJsfBleakClient.values()

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_sf_no_crc(patch_bleak_client) -> None:
    """Test E&J technology BMS data update with no CRC."""

    patch_bleak_client(MockEJsfnoCRCBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "libattU_MockBLEDevice"), True)

    assert await bms.async_update() == MockEJsfnoCRCBleakClient.values()

    await bms.disconnect()


async def test_invalid(patch_bleak_client) -> None:
    """Test E&J technology BMS data update."""

    patch_bleak_client(MockEJinvalidBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == {}
    await bms.disconnect()


@pytest.fixture(
    name="wrong_response",
    params=[
        (b"x009031001E0000001400080016F4~", "wrong SOI"),
        (b":009031001E0000001400080016F4x", "wrong EOI"),
        (b":009031001D0000001400080016F4~", "wrong length"),
        (b":009031001E00000002000A000AD9~", "wrong CRC"),
        (b":009031001E000X001400080016F4~", "wrong encoding"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request: pytest.FixtureRequest) -> bytes:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytes)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockEJBleakClient, "_response", lambda _s, _c, _d: wrong_response
    )

    patch_bleak_client(MockEJBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b":008231008C000000000000000CBF0CC00CEA0CD50000000000000000000000000000000000000000"
                b"00000000008C000041282828F004000000000100004B044C05DC05DCAE~"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b":008231008C000000000000000CBF0CC00CEA0CD50000000000000000000000000000000000000000"
                b"00000000008C000041282828F800000000000100004B044C05DC05DCAA~"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(
        MockEJBleakClient, "_response", lambda _s, _c, _d: problem_response[0]
    )

    patch_bleak_client(MockEJBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result.get("problem", False)  # we expect a problem
    assert result.get("problem_code", 0) == (
        0x4 if problem_response[1] == "first_bit" else 0x800
    )

    await bms.disconnect()



"""Test the Super-B BMS implementation."""

from collections.abc import Awaitable, Callable
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.superb_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 32

_PROTO_DEFS: Final[bytearray] = bytearray(
    b"\x00\x75\x5e\x64\x00\x00\x01\xa4\xbe\xcc\xcc\xcd\x41\x62\x89\xc5\x00\x00\x00\x00"
)

_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 14.159,
    "current": -0.4,
    "battery_health": 100,
    "battery_level": 94,
    "power": -5.664,
    "runtime": 420,
    "battery_charging": False,
    "balancer": False,
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockSuperBBleakClient(MockBleakClient):
    """Emulate a Super-B BMS BleakClient."""

    _RESP: bytearray = _PROTO_DEFS

    def _send_info(self) -> None:
        assert self._notify_callback is not None
        for notify_data in [
            self._RESP[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._RESP), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockSuperBBleakClient", notify_data)

    @property
    def is_connected(self) -> bool:
        """Mock connected."""
        if self._connected:
            self._send_info()  # patch to provide data when not reconnecting
        return self._connected

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs,
    ) -> None:
        """Mock start_notify."""
        await super().start_notify(char_specifier, callback)
        self._send_info()


async def test_update(monkeypatch: pytest.MonkeyPatch, patch_bleak_client, keep_alive_fixture) -> None:
    """Test Super-B BMS data update."""

    monkeypatch.setattr(MockSuperBBleakClient, "_RESP", _PROTO_DEFS)
    patch_bleak_client(MockSuperBBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_chrg(monkeypatch: pytest.MonkeyPatch, patch_bleak_client) -> None:
    """Test Super-B BMS data update with positive current (charging)."""

    monkeypatch.setattr(
        MockSuperBBleakClient,
        "_RESP",
        bytearray(
            b"\x00\x75\x5e\x64\x00\x00\x01\xa4\x3e\xcc\xcc\xcd\x41\x62\x89\xc5\x00\x00\x00\x00"
        ),
    )
    patch_bleak_client(MockSuperBBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = _RESULT_DEFS.copy() | {
        "current": 0.4,
        "battery_charging": True,
        "power": 5.664,
    }
    del result["runtime"]
    assert await bms.async_update() == result


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockSuperBBleakClient, BMS)


async def test_tx_notimplemented(patch_bleak_client) -> None:
    """Test Super-B BMS uuid_tx not implemented for coverage."""

    patch_bleak_client(MockSuperBBleakClient)

    bms = BMS(generate_ble_device(), False)

    with pytest.raises(NotImplementedError):
        _ret = bms.uuid_tx()


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"",
        b"\x00\x72\x5e\x64\x00\x00\x01\xa4\xbe\xcc\xcc\xcd\x41\x62\x89\xc5\x00\x00\x00",
    ],
    ids=["empty", "too_short"],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout, wrong_response: bytes
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout("superb_bms")
    monkeypatch.setattr(MockSuperBBleakClient, "_RESP", bytearray(wrong_response))
    patch_bleak_client(MockSuperBBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(
    ("problem_response"),
    [
        b"\x00\x74\x5e\x64\x00\x00\x01\xa4\xbe\xcc\xcc\xcd\x41\x62\x89\xc5\x00\x00\x00\x00",
        b"\x00\x72\x5e\x64\x00\x00\x01\xa4\xbe\xcc\xcc\xcd\x41\x62\x89\xc5\x00\x00\x00\x00",
    ],
    ids=["chrg_warning", "dischrg_warning"],
)
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, problem_response
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(MockSuperBBleakClient, "_RESP", bytearray(problem_response))

    patch_bleak_client(MockSuperBBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == _RESULT_DEFS | {"problem": True, "problem_code": 1}

    await bms.disconnect()



"""Test the Topband BMS implementation."""

from collections.abc import Awaitable, Callable
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.topband_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 32

_PROTO_DEFS: Final[dict[int, bytearray]] = {
    0x5E: bytearray(  # Ective
        b"\x36\x46\x32\x00\x5e\x38\x34\x33\x35\x30\x30\x30\x30\x46\x38\x43\x44\x46\x46\x46\x46"
        b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30\x45\x31\x30\x42\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45"
        b"\x30\x44\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x39\x34\x46\xaf\x46\x38\x33\x33\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x00\x00\x00\x00\x00\x00\x00\x00"  # \xaf ... garbage
    ),
    0x83: bytearray(  # StartCraft
        b"\x83\x36\x32\x33\x34\x30\x30\x30\x30\x37\x36\x46\x45\x46\x46\x46\x46\x38\x38\x38\x41"
        b"\x30\x31\x30\x30\x31\x36\x30\x30\x36\x32\x30\x30\x35\x33\x30\x42\x30\x30\x38\x30\x30"
        b"\x37\x42\x34\x31\x36\x30\x44\x31\x37\x30\x44\x31\x39\x30\x44\x31\x43\x30\x44\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x37\x43\x32\x11\x11\x11\x11\x11\x11\x11\x11"
    ),
    0xB0: bytearray(  # KiloVault
        b"\xb0\x39\x30\x33\x33\x30\x30\x30\x30\x33\x45\x46\x45\x46\x46\x46\x46\x33\x34\x30\x46"
        b"\x30\x33\x30\x30\x37\x44\x30\x30\x35\x45\x30\x30\x39\x31\x30\x42\x30\x30\x30\x31\x30"
        b"\x30\x30\x30\x44\x45\x30\x43\x30\x38\x30\x44\x45\x32\x30\x43\x30\x35\x30\x44\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
        b"\x30\x30\x30\x30\x30\x37\x42\x41\x52\x52\x52\x52\x52\x52\x52\x52"
    ),
}

_RESULT_DEFS: Final[dict[int, BMSSample]] = {
    0x5E: {
        "voltage": 13.7,
        "current": -12.808,
        "battery_level": 98,
        "cycles": 407,
        "cycle_charge": 194.86,
        "cell_count": 4,
        "cell_voltages": [3.422, 3.441, 3.429, 3.422],
        "delta_voltage": 0.019,
        "temperature": 30.95,
        "temp_values": [30.95],
        "cycle_capacity": 2669.582,
        "power": -175.47,
        "runtime": 54770,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
    },
    0x83: {
        "voltage": 13.41,
        "current": -0.394,
        "battery_level": 98,
        "cycles": 22,
        "cycle_charge": 101,
        "cell_count": 4,
        "cell_voltages": [3.35, 3.351, 3.353, 3.356],
        "delta_voltage": 0.006,
        "temperature": 16.75,
        "temp_values": [16.75],
        "cycle_capacity": 1354.41,
        "power": -5.284,
        "runtime": 922842,
        "battery_charging": False,
        "problem": False,
        "problem_code": 0,
    },
    0xB0: {
        "battery_charging": False,
        "cycle_capacity": 2646.6,
        "voltage": 13.2,
        "current": -0.45,
        "cycle_charge": 200.5,
        "delta_voltage": 0.042,
        "power": -5.94,
        "problem": False,
        "cycles": 125,
        "battery_level": 94,
        "temperature": 22.95,
        "temp_values": [22.95],
        "problem_code": 0,
        "runtime": 1604000,
        "cell_count": 4,
        "cell_voltages": [3.294, 3.336, 3.298, 3.333],
    },
}


@pytest.fixture(
    name="protocol_type",
    params=[0x5E, 0x83, 0xB0],
)
def proto(request: pytest.FixtureRequest) -> int:
    """Protocol fixture."""
    assert isinstance(request.param, int)
    return request.param


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS

class MockTopbandBleakClient(MockBleakClient):
    """Emulate a Topband BMS BleakClient."""

    _RESP: bytearray = _PROTO_DEFS[0x5E]

    def _send_info(self) -> None:
        assert self._notify_callback is not None
        for notify_data in [
            self._RESP[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._RESP), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockTopbandBleakClient", notify_data)

    @property
    def is_connected(self) -> bool:
        """Mock connected."""
        if self._connected:
            self._send_info()  # patch to provide data when not reconnecting
        return self._connected

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs,
    ) -> None:
        """Mock start_notify."""
        await super().start_notify(char_specifier, callback)
        self._send_info()


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: int,
    keep_alive_fixture: bool,
) -> None:
    """Test Topband BMS data update."""

    monkeypatch.setattr(MockTopbandBleakClient, "_RESP", _PROTO_DEFS[protocol_type])
    patch_bleak_client(MockTopbandBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockTopbandBleakClient, BMS)


async def test_tx_notimplemented(patch_bleak_client) -> None:
    """Test Topband BMS uuid_tx not implemented for coverage."""

    patch_bleak_client(MockTopbandBleakClient)

    bms = BMS(generate_ble_device(), False)

    with pytest.raises(NotImplementedError):
        _ret = bms.uuid_tx()


@pytest.fixture(
    name="wrong_response",
    params=[
        (
            b"\x5e\x38\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
            b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
            b"\x45\x31\x30\x42\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x38\x38\x45\x00\x00\x00\x00\x00\x00\x00\x00",
            "wrong_CRC",
        ),
        (
            b"\x5a\x38\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
            b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
            b"\x45\x31\x30\x42\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x38\x38\x46\x00\x00\x00\x00\x00\x00\x00\x00",
            "wrong_SOF",
        ),
        (
            b"\x5e\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
            b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
            b"\x45\x31\x30\x42\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x38\x38\x46",
            "wrong_length",  # 1st byte missing
        ),
        (
            b"\x5e\x5e\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
            b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
            b"\x45\x31\x30\x42\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
            b"\x30\x38\x38\x46",
            "wrong_character",
        ),
    ],
    ids=lambda param: param[1],
)
def response(request: pytest.FixtureRequest) -> bytes:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytes)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout("topband_bms")
    monkeypatch.setattr(MockTopbandBleakClient, "_RESP", bytearray(wrong_response))
    patch_bleak_client(MockTopbandBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.fixture(
    name="problem_response",
    params=[
        (
            bytearray(
                b"\x5e\x38\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
                b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
                b"\x45\x31\x30\x42\x30\x31\x30\x30\x30\x30\x30\x30"
                b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x38\x39\x30"
            ),
            "first_bit",
        ),
        (
            bytearray(
                b"\x5e\x38\x34\x33\x35\x30\x30\x30\x30\x33\x38\x43\x44\x46\x46\x46\x46"
                b"\x32\x43\x46\x39\x30\x32\x30\x30\x39\x37\x30\x31\x36\x32\x30\x30"
                b"\x45\x31\x30\x42\x38\x30\x30\x30\x30\x30\x30\x30"
                b"\x35\x45\x30\x44\x37\x31\x30\x44\x36\x35\x30\x44\x35\x45\x30\x44"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x39\x30\x46"
            ),
            "last_bit",
        ),
    ],
    ids=lambda param: param[1],
)
def prb_response(request) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning error flags."""

    monkeypatch.setattr(MockTopbandBleakClient, "_RESP", bytearray(problem_response[0]))

    patch_bleak_client(MockTopbandBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == {
        "voltage": 13.7,
        "current": -13.0,
        "battery_level": 98,
        "cycles": 407,
        "cycle_charge": 194.86,
        "cell_count": 4,
        "cell_voltages": [
            3.422,
            3.441,
            3.429,
            3.422,
        ],
        "delta_voltage": 0.019,
        "temperature": 30.95,
        "temp_values": [30.95],
        "cycle_capacity": 2669.582,
        "power": -178.1,
        "runtime": 53961,
        "battery_charging": False,
        "problem": True,
        "problem_code": (1 if problem_response[1] == "first_bit" else 128),
    }

    await bms.disconnect()



"""Test the Jikong BMS implementation."""

import asyncio
from collections.abc import Buffer
from copy import deepcopy
from typing import Final, Literal, cast
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSInfo, BMSMode, BMSSample
from aiobmsble.basebms import crc_sum, lstr2int
from aiobmsble.bms.jikong_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import DefGATTChar, MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 128

_PROTO_DEFS: Final[dict[str, dict[str, bytearray]]] = {
    "JK02_24S": {
        "dev": bytearray(  # JK02_24S (SW: 10.08)
            b"\x55\xaa\xeb\x90\x03\x79\x4a\x4b\x2d\x42\x32\x41\x32\x30\x53\x32\x30\x50\x00\x00\x00"
            b"\x00\x31\x30\x2e\x58\x47\x00\x00\x00\x31\x30\x2e\x30\x38\x00\x00\x00\xe4\xe7\x6c\x03"
            b"\x11\x00\x00\x00\x4a\x4b\x2d\x42\x4d\x53\x2d\x41\x00\x00\x00\x00\x00\x00\x00\x00\x31"
            b"\x32\x33\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x32\x30\x37\x30\x31"
            b"\x00\x00\x32\x30\x33\x32\x38\x31\x36\x30\x31\x32\x00\x30\x30\x30\x30\x00\x4d\x61\x72"
            b"\x69\x6f\x00\x00\x00\x00\x00\x00\x00\x00\x61\x00\x00\x31\x32\x33\x34\x35\x36\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x93"
        ),
        "ack": bytearray(
            b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x41"
            b"\x54\x0d\x0a"
        ),  # ACKnowledge message with attached AT\r\n message (needs to be filtered)
        "cell": bytearray(  # JK02_24S (SW: 10.08)
            b"\x55\xaa\xeb\x90\x02\xc8\xee\x0c\xf2\x0c\xf1\x0c\xf0\x0c\xf0\x0c\xec\x0c\xf0\x0c\xed"
            b"\x0c\xed\x0c\xed\x0c\xed\x0c\xf0\x0c\xf1\x0c\xed\x0c\xee\x0c\xed\x0c\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\x00\xef\x0c\x05\x00\x01"
            b"\x09\x36\x00\x37\x00\x39\x00\x38\x00\x37\x00\x37\x00\x35\x00\x41\x00\x42\x00\x36\x00"
            b"\x37\x00\x3a\x00\x38\x00\x34\x00\x36\x00\x37\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xeb\xce\x00\x00\xc7\x0d\x02\x00"
            b"\x19\x09\x00\x00\xb5\x00\xba\x00\xe4\x00\x00\x00\x02\x00\x00\x38\x5d\xba\x01\x00\x10"
            b"\x15\x03\x00\x3c\x00\x00\x00\xa4\x65\xb9\x00\x64\x00\xd9\x02\x8b\xe8\x6c\x03\x01\x01"
            b"\xb3\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00\x01\x00\x00\x00\x23"
            b"\x04\x0b\x00\x00\x00\x9f\x19\x40\x40\x00\x00\x00\x00\xe2\x04\x00\x00\x00\x00\x00\x01"
            b"\x00\x03\x00\x00\x83\xd5\x37\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\xbd"
        ),
    },
    "JK02_32S": {  # JK02_32 (SW: V11.48)
        "dev": bytearray(
            b"\x55\xaa\xeb\x90\x03\xa3\x4a\x4b\x5f\x42\x32\x41\x38\x53\x32\x30\x50\x00\x00\x00\x00"
            b"\x00\x31\x31\x2e\x58\x41\x00\x00\x00\x31\x31\x2e\x34\x38\x00\x00\x00\xe4\xa7\x46\x00"
            b"\x07\x00\x00\x00\x31\x32\x76\x34\x32\x30\x61\x00\x00\x00\x00\x00\x00\x00\x00\x00\x31"
            b"\x32\x33\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x34\x30\x37\x30\x34"
            b"\x00\x00\x34\x30\x34\x30\x39\x32\x43\x32\x32\x36\x32\x00\x30\x30\x30\x00\x49\x6e\x70"
            b"\x75\x74\x20\x55\x73\x65\x72\x64\x61\x74\x61\x00\x00\x31\x34\x30\x37\x30\x33\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\xfe\xf9\xff\xff\x1f\x2d\x00\x02\x00\x00\x00\x00\x90\x1f\x00\x00\x00\x00"
            b"\xc0\xd8\xe7\x32\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x41\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x09\x00\x00\x00\x64\x00\x00\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\xbf\x21\x06\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\xd8"
        ),  # Vendor_ID: JK_B2A8S20P, SN: 404092C2262, HW: V11.XA, SW: V11.48, power-on: 7, Version: 4.28.0
        "ack": bytearray(
            b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x41"
            b"\x54\x0d\x0a"
        ),  # ACKnowledge message with attached AT\r\n message (needs to be filtered)
        "cell": bytearray(
            b"\x55\xaa\xeb\x90\x02\xad\xf3\x0c\xf3\x0c\xf3\x0c\xf0\x0c\xf1\x0c\xf0\x0c\xf1\x0c\xf1"
            b"\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00\x00\xf2\x0c\x03\x00\x00\x07\x38\x00\x37\x00"
            b"\x36\x00\x37\x00\x36\x00\x37\x00\x36\x00\x37\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x36\x01\x00"
            b"\x00\x00\x00\x8d\x67\x00\x00\x60\xdb\x02\x00\x69\xe4\xff\xff\x1c\x01\x24\x01\x00\x00"
            b"\x00\x00\x00\x00\x00\x44\x80\x2c\x02\x00\x50\x34\x03\x00\x15\x00\x00\x00\xbc\x62\x44"
            b"\x00\x64\x00\x00\x00\x1e\xf3\x68\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\xff\x00\x01\x00\x00\x00\xf1\x03\x00\x00\x23\x00\x29\xb4\x3f\x40\x00"
            b"\x00\x00\x00\x5a\x0a\x00\x00\x00\x01\x00\x01\x00\x06\x00\x00\xef\x3d\x08\x04\x00\x00"
            b"\x00\x00\x36\x01\x00\x00\x00\x00\xf1\x03\x64\x39\x67\x00\x1a\x00\x00\x00\x80\x51\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xfe\xff\x7f\xdc\x0f\x01\x00"
            b"\x80\x03\x00\x00\x00\xb4"
        ),
    },
    "JK02_32S_v15": {  # JK02_32 (SW: V15.38)
        "dev": bytearray(
            b"\x55\xaa\xeb\x90\x03\x21\x4a\x4b\x5f\x50\x42\x32\x41\x31\x36\x53\x32\x30\x50\x00\x00"
            b"\x00\x31\x35\x41\x00\x00\x00\x00\x00\x31\x35\x2e\x33\x38\x00\x00\x00\x20\x48\x01\x00"
            b"\x05\x00\x00\x00\x34\x31\x30\x31\x38\x34\x39\x32\x35\x35\x35\x00\x50\x00\x00\x00\x31"
            b"\x32\x33\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x35\x30\x32\x31\x30"
            b"\x00\x00\x34\x31\x30\x31\x38\x34\x39\x32\x35\x35\x35\x00\x30\x30\x30\x00\x4a\x4b\x2d"
            b"\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x37\x31\x32\x30\x33\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x4a\x4b\x2d\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\xfe\xff\xff\xff\x8f\xe9\x1d\x02\x00\x00\x00\x00\x90\x1f\x00\x00\x00\x00"
            b"\xc0\xd8\xe7\xfe\x3f\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\xff\x67\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x0f\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x01\xff\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x09\x08\x00\x01\x64\x00\x00\x00\x5f\x00\x00\x00\x3c\x00\x00\x00\x32\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x0e\x00\x00\x0a\x50\x01\x1e\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x9f\xe9\xff\x0f\x00\x00"
            b"\x00\x00\x00\x00\x00\xf1"
        ),
        "ack": bytearray(
            b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x41"
            b"\x54\x0d\x0a"
        ),  # ACKnowledge message with attached AT\r\n message (needs to be filtered)
        "cell": bytearray(
            b"\x55\xaa\xeb\x90\x02\xac\x05\x0d\xfe\x0c\xfe\x0c\x01\x0d\x01\x0d\xfd\x0c\xfb\x0c\x01"
            b"\x0d\xfc\x0c\xfb\x0c\xfe\x0c\xfb\x0c\xf8\x0c\xfb\x0c\xfb\x0c\x09\x0d\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\x00\xff\x0c\x10\x00\x0f\x0c\x40\x00\x3d\x00"
            b"\x40\x00\x3d\x00\x41\x00\x3f\x00\x41\x00\x3e\x00\x41\x00\x3e\x00\x41\x00\x3d\x00\x40"
            b"\x00\x3e\x00\x41\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x00\x00"
            b"\x00\x00\x00\xe8\xcf\x00\x00\x4a\xe4\x19\x00\x89\x7c\x00\x00\x86\x00\x80\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x19\x86\xc0\x00\x00\x40\x0d\x03\x00\x09\x00\x00\x00\xb1\x5f\x1c"
            b"\x00\x64\x00\x00\x00\x8c\x4c\x76\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\xff\x00\x01\x00\x00\x00\x05\x04\x52\x00\x00\x00\xd4\x4e\x40\x40\x00"
            b"\x00\x00\x00\xca\x14\x00\x00\x00\x01\x01\x01\x00\x06\x00\x00\xea\x67\x00\x00\x00\x00"
            b"\x00\x00\xcd\x00\xc3\x00\xbf\x00\xbd\x03\x28\x7f\x03\x0a\x1d\x00\x00\x00\x80\x51\x01"
            b"\x00\x00\x00\x01\x00\xa5\x02\x02\x00\x00\x00\x00\x00\x00\xfe\xff\x7f\xdc\x2f\x01\x01"
            b"\xb0\xcf\x07\x00\x00\xbb"
        ),  # 53.224V, 25%, 31.881A, cycles:9, 16.367°C, Float, timer: 677s,
    },
    "JK02_32S_v19.05": {  # JK02_32 (SW: V19.05)
        "dev": bytearray(
            b"\x55\xaa\xeb\x90\x03\x98\x4a\x4b\x5f\x50\x42\x32\x41\x31\x36\x53\x32\x30\x50\x00\x00"
            b"\x00\x31\x39\x41\x00\x00\x00\x00\x00\x31\x39\x2e\x30\x35\x00\x00\x00\x48\x73\x08\x00"
            b"\x0b\x00\x00\x00\x42\x61\x74\x65\x72\x69\x65\x20\x31\x00\x00\x00\x00\x00\x00\x00\x31"
            b"\x32\x33\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x35\x30\x35\x32\x34"
            b"\x00\x00\x35\x30\x33\x32\x31\x34\x38\x34\x39\x30\x30\x30\x36\x34\x33\x00\x4a\x4b\x2d"
            b"\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x33\x31\x34\x31\x35\x39\x32\x37"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x4a\x4b\x2d\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\xff\xff\xff\xff\x8f\xe9\x8d\x03\x00\x00\x00\x00\x90\x1f\x00\x00\x00\x00"
            b"\xc0\xd8\xf7\xfe\x7f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x04\xff\xe7\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x0f\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x09\x08\x00\x01\x64\x00\x00\x00\x5f\x00\x00\x00\x3c\x00\x00\x00\x32\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x0e\x00\x00\x0a\x3c\x01\x1e\x0f\x03\xa4"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x9f\xe9\xfe\x1b\x00\x00"
            b"\x00\x00\x00\x00\x00\xc5\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x44\x41\x54\x0d\x0a"
        ),  # Vendor_ID: JK_B2A16S20P, SN: 50321489000643, HW: V19.XA, SW: V19.05, power-on: 11
        "ack": bytearray(),  # ACKnowledge is contained at end of dev msg, with AT cmd
        "cell": bytearray(  # copy from v15
            b"\x55\xaa\xeb\x90\x02\xac\x05\x0d\xfe\x0c\xfe\x0c\x01\x0d\x01\x0d\xfd\x0c\xfb\x0c\x01"
            b"\x0d\xfc\x0c\xfb\x0c\xfe\x0c\xfb\x0c\xf8\x0c\xfb\x0c\xfb\x0c\x09\x0d\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\x00\xff\x0c\x10\x00\x0f\x0c\x40\x00\x3d\x00"
            b"\x40\x00\x3d\x00\x41\x00\x3f\x00\x41\x00\x3e\x00\x41\x00\x3e\x00\x41\x00\x3d\x00\x40"
            b"\x00\x3e\x00\x41\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x00\x00"
            b"\x00\x00\x00\xe8\xcf\x00\x00\x4a\xe4\x19\x00\x89\x7c\x00\x00\x86\x00\x80\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x19\x86\xc0\x00\x00\x40\x0d\x03\x00\x09\x00\x00\x00\xb1\x5f\x1c"
            b"\x00\x64\x00\x00\x00\x8c\x4c\x76\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\xff\x00\x01\x00\x00\x00\x05\x04\x52\x00\x00\x00\xd4\x4e\x40\x40\x00"
            b"\x00\x00\x00\xca\x14\x00\x00\x00\x01\x01\x01\x00\x06\x00\x00\xea\x67\x00\x00\x00\x00"
            b"\x00\x00\xcd\x00\xc3\x00\xbf\x00\xbd\x03\x28\x7f\x03\x0a\x1d\x00\x00\x00\x80\x51\x01"
            b"\x00\x00\x00\x01\x00\xa5\x02\x02\x00\x00\x00\x00\x00\x00\xfe\xff\x7f\xdc\x2f\x01\x01"
            b"\xb0\xcf\x07\x00\x00\xbb"
        ),  # 53.224V, 25%, 31.881A, cycles:9, 16.367°C, Float, timer: 677s,
    },
    "JK02_32S_v19.27": {  # JK02_32 (SW: V19.27)
        "dev": bytearray(
            b"\x55\xaa\xeb\x90\x03\xda\x4a\x4b\x2d\x50\x42\x32\x41\x31\x36\x53\x32\x30\x50\x00\x00"
            b"\x10\x31\x39\x41\x00\x00\x00\x00\x00\x31\x39\x2e\x32\x37\x00\x00\x00\xc0\x2d\x21\x00"
            b"\x6c\x00\x00\x00\x44\x47\x20\x53\x6d\x61\x72\x74\x20\x42\x4d\x53\x00\x00\x00\x00\x31"
            b"\x35\x35\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x35\x31\x32\x32\x31"
            b"\x00\x00\x35\x31\x30\x32\x30\x42\x4f\x34\x39\x30\x30\x30\x34\x32\x32\x00\x4a\x4b\x2d"
            b"\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x39\x32\x32\x37\x34\x36\x30\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x4a\x4b\x2d\x42\x4d\x53\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\xff\xff\xff\xff\xaf\xe9\x8d\x0f\x00\x00\x00\x00\x90\x1f\x00\x00\x00\x00"
            b"\xc0\xd8\xf7\xfe\xef\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0e\x00\xfb\xe6\x00"
            b"\x00\x43\x45\x48\x4d\x50\x52\x54\x00\x00\x00\x00\x00\xff\x0f\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x09\x08\x00\x01\x64\x00\x00\x00\x5f\x00\x00\x00\x3c\x00\x00\x00\x32\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x0e\x00\x00\x05\x32\x01\x1e\x0f\x03\xa4"
            b"\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x9f\xeb\xfe\x19\x00\x00"
            b"\x00\x00\x00\x00\x02\xd2"
        ),
        "ack": bytearray(
            b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44"
        ),
        "cell": bytearray(
            b"\x55\xaa\xeb\x90\x02\xda\xec\x0c\xf0\x0c\xf0\x0c\xeb\x0c\xef\x0c\xef\x0c\xf0\x0c\xed"
            b"\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00\x00\xee\x0c\x05\x00\x01\x03\x61\x00\x5c\x00"
            b"\x5f\x00\x55\x00\x60\x00\x57\x00\x65\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x01\x00"
            b"\x00\x00\x00\x71\x67\x00\x00\xfe\x2c\x05\x00\x74\xce\xff\xff\xe9\x00\xec\x00\x00\x00"
            b"\x00\x00\xc6\x07\x01\x4e\x48\xba\x03\x00\x90\xca\x04\x00\x0f\x00\x00\x00\xe9\x24\x4a"
            b"\x00\x64\x00\x00\x00\x0f\x2e\x21\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\xff\x00\x01\x00\x00\x00\x6f\x24\x00\x00\x67\x00\x1c\xf9\x3e\x40\x00"
            b"\x00\xd8\x00\x58\x0a\x00\x00\x00\x01\x01\x01\x00\x06\x01\x00\x71\x28\x1f\x00\x00\x00"
            b"\x00\x00\x06\x01\xf5\x00\xf0\x00\x01\x00\x92\xb1\x83\x0b\x3b\x04\x00\x00\x80\x51\x01"
            b"\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\xff\x7f\xdd\x2f\x01\x01"
            b"\xb0\x0f\x07\x00\x00\x84"
        ),
    },
}

_RESULT_DEFS: Final[dict[str, BMSSample]] = {
    "JK02_24S": {
        "cell_count": 16,
        "delta_voltage": 0.005,
        "temperature": 19.833,
        "voltage": 52.971,
        "current": 2.329,
        "balance_current": 0.002,
        "battery_level": 56,
        "battery_health": 100,
        "cycle_charge": 113.245,
        "cycles": 60,
        "cell_voltages": [
            3.310,
            3.314,
            3.313,
            3.312,
            3.312,
            3.308,
            3.312,
            3.309,
            3.309,
            3.309,
            3.309,
            3.312,
            3.313,
            3.309,
            3.310,
            3.309,
        ],
        "cycle_capacity": 5998.701,
        "design_capacity": 202,
        "power": 123.369,
        "battery_charging": True,
        "temp_values": [18.1, 18.6, 22.8],
        "temp_sensors": 7,
        "problem": False,
        "problem_code": 0,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
    },
    "JK02_32S": {
        "cell_count": 8,
        "delta_voltage": 0.003,
        "voltage": 26.509,
        "current": -7.063,
        "battery_level": 68,
        "battery_health": 100,
        "cycle_charge": 142.464,
        "design_capacity": 210,
        "cycles": 21,
        "balance_current": 0.0,
        "temp_sensors": 255,
        "problem_code": 0,
        "temp_values": [31.0, 28.4, 29.2, 31.0],
        "cell_voltages": [3.315, 3.315, 3.315, 3.312, 3.313, 3.312, 3.313, 3.313],
        "cycle_capacity": 3776.578,
        "power": -187.233,
        "battery_charging": False,
        "runtime": 72613,
        "temperature": 29.9,
        "problem": False,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
    },
    "JK02_32S_v15": {
        "cell_count": 16,
        "delta_voltage": 0.016,
        "battery_mode": BMSMode.FLOAT,
        "voltage": 53.224,
        "current": 31.881,
        "battery_level": 25,
        "battery_health": 100,
        "cycle_charge": 49.286,
        "design_capacity": 200,
        "cycles": 9,
        "balance_current": 0.0,
        "temp_sensors": 255,
        "problem_code": 0,
        "temp_values": [12.9, 13.4, 12.8, 20.5, 19.5, 19.1],
        "cell_voltages": [
            3.333,
            3.326,
            3.326,
            3.329,
            3.329,
            3.325,
            3.323,
            3.329,
            3.324,
            3.323,
            3.326,
            3.323,
            3.32,
            3.323,
            3.323,
            3.337,
        ],
        "cycle_capacity": 2623.198,
        "power": 1696.834,
        "battery_charging": True,
        "temperature": 16.367,
        "problem": False,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
    },
    "JK02_32S_v19.05": {
        "cell_count": 16,
        "delta_voltage": 0.016,
        "battery_mode": BMSMode.FLOAT,
        "voltage": 53.224,
        "current": 31.881,
        "battery_level": 25,
        "battery_health": 100,
        "cycle_charge": 49.286,
        "design_capacity": 200,
        "cycles": 9,
        "balance_current": 0.0,
        "temp_sensors": 255,
        "problem_code": 0,
        "temp_values": [12.9, 13.4, 12.8, 20.5, 19.5, 19.1],
        "cell_voltages": [
            3.333,
            3.326,
            3.326,
            3.329,
            3.329,
            3.325,
            3.323,
            3.329,
            3.324,
            3.323,
            3.326,
            3.323,
            3.32,
            3.323,
            3.323,
            3.337,
        ],
        "cycle_capacity": 2623.198,
        "power": 1696.834,
        "battery_charging": True,
        "temperature": 16.367,
        "problem": False,
        "balancer": False,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
    },
    "JK02_32S_v19.27": {
        "voltage": 26.481,
        "current": -12.684,
        "problem_code": 0,
        "balance_current": 1.99,
        "balancer": True,
        "battery_level": 78,
        "cycle_charge": 244.296,
        "design_capacity": 314,
        "cycles": 15,
        "battery_health": 100,
        "chrg_mosfet": True,
        "dischrg_mosfet": True,
        "temp_sensors": 255,
        "cell_count": 8,
        "delta_voltage": 0.005,
        "battery_mode": BMSMode.BULK,
        "temp_values": [26.2, 23.3, 23.6, 26.2, 24.5, 24.0],
        "cell_voltages": [3.308, 3.312, 3.312, 3.307, 3.311, 3.311, 3.312, 3.309],
        "cycle_capacity": 6469.202,
        "power": -335.885,
        "battery_charging": False,
        "runtime": 69336,
        "temperature": 24.633,
        "problem": False,
    },
}

_DEV_DEFS: Final[dict[str, BMSInfo]] = {
    "JK02_24S": {
        "model": "JK-B2A20S20P",
        "hw_version": "10.XG",
        "sw_version": "10.08",
        "name": "JK-BMS-A",
        "serial_number": "20328160",
    },
    "JK02_32S": {
        "model": "JK_B2A8S20P",
        "hw_version": "11.XA",
        "sw_version": "11.48",
        "name": "12v420a",
        "serial_number": "404092C2",
    },
    "JK02_32S_v15": {
        "model": "JK_PB2A16S20P",
        "hw_version": "15A",
        "sw_version": "15.38",
        "name": "41018492555",
        "serial_number": "41018492",
    },
    "JK02_32S_v19.05": {
        "model": "JK_PB2A16S20P",
        "hw_version": "19A",
        "sw_version": "19.05",
        "name": "Baterie 1",
        "serial_number": "50321484",
    },
    "JK02_32S_v19.27": {
        "model": "JK-PB2A16S20P",
        "hw_version": "19A",
        "sw_version": "19.27",
        "name": "DG Smart BMS",
        "serial_number": "51020BO4",
    },
}


@pytest.fixture(
    name="protocol_type",
    params=[
        "JK02_24S",
        "JK02_32S",
        "JK02_32S_v15",
        "JK02_32S_v19.05",
        "JK02_32S_v19.27",
    ],
)
def proto(request: pytest.FixtureRequest) -> str:
    """Protocol fixture."""
    assert isinstance(request.param, str)
    return request.param


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockJikongBleakClient(MockBleakClient):
    """Emulate a Jikong BMS BleakClient."""

    HEAD_CMD: Final = bytearray(b"\xaa\x55\x90\xeb")
    CMD_INFO: Final = bytearray(b"\x96")
    DEV_INFO: Final = bytearray(b"\x97")
    _FRAME: dict[str, bytearray] = {}

    _task: asyncio.Task[None] | None = None

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        frame: Final[bytearray] = bytearray(data)
        if char_specifier != 3 or crc_sum(frame[:-1]) != frame[19]:
            return bytearray()

        if frame.startswith(self.HEAD_CMD + self.CMD_INFO):
            return (
                bytearray(b"\x41\x54\x0d\x0a") + self._FRAME["cell"]
            )  # added AT\r\n command
        if frame.startswith(self.HEAD_CMD + self.DEV_INFO):
            return self._FRAME["dev"]

        return bytearray()

    async def _send_confirm(self) -> None:
        assert self._notify_callback, "send confirm called but notification not enabled"
        await asyncio.sleep(0)
        self._notify_callback(
            "MockJikongBleakClient", self._FRAME.get("ack", bytearray())
        )

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"
        self._notify_callback(
            "MockJikongBleakClient", bytearray(b"\x41\x54\x0d\x0a")
        )  # interleaved AT\r\n command
        resp = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockJikongBleakClient", notify_data)
        if bytes(data).startswith(
            self.HEAD_CMD + self.DEV_INFO
        ):  # JK BMS confirms commands with a command in reply
            self._task = asyncio.create_task(self._send_confirm())

    async def disconnect(self) -> None:
        """Mock disconnect and wait for send task."""
        if self._task is not None:
            await asyncio.wait_for(self._task, 0.1)
            assert self._task.done(), "send task still running!"
        await super().disconnect()

    @property
    def services(self) -> BleakGATTServiceCollection:
        """Emulate JiKong BT service setup."""

        serv_col = BleakGATTServiceCollection()
        service = BleakGATTService(None, 1, uuid=normalize_uuid_str("ffe0"))
        service.add_characteristic(
            DefGATTChar(
                service.handle + 1,
                uuid=f"{service.uuid[4:7]!s}1",
                properties=["notify"],
                service=service,
            )
        )
        service.add_characteristic(
            DefGATTChar(
                service.handle + 2,
                uuid=f"{service.uuid[4:7]!s}1",
                properties=["write", "write-without-response"],
                service=service,
            )
        )
        service.add_characteristic(
            DefGATTChar(
                service.handle + 3,
                uuid="0000",
                properties=["write", "write-without-response"],
                service=service,
            )
        )

        serv_col.add_service(service)

        return serv_col


class MockStreamBleakClient(MockJikongBleakClient):
    """Mock JiKong BMS that already sends battery data (no request required)."""

    async def _send_all(self) -> None:
        assert (
            self._notify_callback
        ), "send_all frames called but notification not enabled"
        for resp in self._FRAME.values():
            self._notify_callback("MockJikongBleakClient", resp)
            await asyncio.sleep(0)


class MockWrongBleakClient(MockBleakClient):
    """Mock invalid service for JiKong BMS."""

    @property
    def services(self) -> BleakGATTServiceCollection:
        """Emulate JiKong BT service setup."""

        return BleakGATTServiceCollection()


class MockInvalidBleakClient(MockJikongBleakClient):
    """Emulate a Jikong BMS BleakClient with disconnect error."""

    async def disconnect(self) -> None:
        """Mock disconnect to raise BleakError."""
        if self._task is not None:
            await asyncio.wait_for(self._task, 0.1)
            assert self._task.done(), "send task still running!"
        raise BleakError


class MockOversizedBleakClient(MockJikongBleakClient):
    """Emulate a Jikong BMS BleakClient returning wrong data length."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if char_specifier != 3:
            return bytearray()
        if bytearray(data)[0:5] == self.HEAD_CMD + self.CMD_INFO:
            return (  # added AT\r\n command and oversized
                bytearray(b"\x41\x54\x0d\x0a") + self._FRAME["cell"] + bytearray(6)
            )
        if bytearray(data)[0:5] == self.HEAD_CMD + self.DEV_INFO:
            return self._FRAME["dev"] + bytearray(6)  # oversized

        return bytearray()


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: str,
    keep_alive_fixture,
) -> None:
    """Test Jikong BMS data update."""

    monkeypatch.setattr(MockJikongBleakClient, "_FRAME", _PROTO_DEFS[protocol_type])

    patch_bleak_client(MockJikongBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]

    # query again to check already connected state
    assert await bms.async_update() == _RESULT_DEFS[protocol_type]
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, protocol_type: str
) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    monkeypatch.setattr(MockJikongBleakClient, "_FRAME", _PROTO_DEFS[protocol_type])
    patch_bleak_client(MockJikongBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == _DEV_DEFS[protocol_type]


async def test_hide_temp_sensors(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, protocol_type: str
) -> None:
    """Test Jikong BMS data update with not connected temperature sensors."""

    temp12_hide: dict[str, bytearray] = deepcopy(_PROTO_DEFS[protocol_type])

    # clear temp sensor #2
    if protocol_type == "JK02_24S":
        temp12_hide["cell"][182:184] = bytearray(b"\x03\x00")
        temp12_hide["cell"][132:134] = bytearray(b"\x30\xf8")  # -200.0
    else:
        temp12_hide["cell"][214:216] = bytearray(b"\xfb\x00")
        temp12_hide["cell"][162:164] = bytearray(b"\x30\xf8")  # -200.0
    # recalculate CRC
    temp12_hide["cell"][-1] = crc_sum(temp12_hide["cell"][:-1])

    monkeypatch.setattr(MockJikongBleakClient, "_FRAME", temp12_hide)

    patch_bleak_client(MockJikongBleakClient)

    bms = BMS(generate_ble_device())

    # modify result dict to match removed temp#1, temp#2
    ref_result: BMSSample = deepcopy(_RESULT_DEFS[protocol_type])
    if protocol_type == "JK02_24S":
        ref_result |= {"temp_sensors": 3, "temperature": 18.1}
    elif protocol_type == "JK02_32S":
        ref_result |= {"temp_sensors": 251, "temperature": 31.0}
    elif protocol_type in ("JK02_32S_v15", "JK02_32S_v19.05"):
        ref_result |= {"temp_sensors": 251, "temperature": 18.0}
    elif protocol_type == "JK02_32S_v19.27":
        ref_result |= {"temp_sensors": 251, "temperature": 25.225}

    temp_values: list[int | float] = ref_result.get("temp_values", [])
    temp_values.pop(1)  # remove sensor 1
    temp_values.pop(1)  # remove sensor 2
    ref_result["temp_values"] = temp_values.copy()

    assert await bms.async_update() == ref_result

    await bms.disconnect()


async def test_stream_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: str,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test Jikong BMS data update."""

    _frames: dict[str, bytearray] = deepcopy(_PROTO_DEFS[protocol_type])
    _prot_offset: Literal[-32, 0] = (
        -32 if lstr2int(_frames["dev"][30:38].decode()) < 11 else 0
    )

    monkeypatch.setattr(MockStreamBleakClient, "_FRAME", _frames)
    patch_bleak_client(MockStreamBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]
    assert bms._msg_event.is_set() is False, "BMS does not request fresh data"
    assert "requesting cell info" in caplog.text

    _client = cast(MockStreamBleakClient, bms._client)
    _cell_frame: bytearray = _frames["cell"]
    _cell_frame[_prot_offset + 182] = 0x73  # modify data to ensure new read
    _cell_frame[-1] = crc_sum(_cell_frame[:-1])
    caplog.clear()
    await _client._send_all()

    # query again to see if updated streaming data is used
    assert await bms.async_update() == _RESULT_DEFS[protocol_type] | {"cycles": 0x73}
    assert "requesting cell info" not in caplog.text, "BMS did not use streaming data"

    _cell_frame[_prot_offset + 182] = 0x37  # modify data to ensure new read
    _cell_frame[-1] = crc_sum(_cell_frame[:-1])
    caplog.clear()
    # do not automatically send data this time

    # query again to see if BMS recovers if no data is sent
    assert await bms.async_update() == _RESULT_DEFS[protocol_type] | {"cycles": 0x37}
    assert "requesting cell info" in caplog.text, "BMS did not use streaming data"


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()

    # return type 0x03 (first requested message) with incorrect CRC
    monkeypatch.setattr(
        MockInvalidBleakClient,
        "_response",
        lambda _s, _c, _d: bytearray(b"\x55\xaa\xeb\x90\x03") + bytearray(295),
    )

    patch_bleak_client(MockInvalidBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()
    assert not result

    await bms.disconnect()


async def test_invalid_frame_type(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bms_timeout()

    monkeypatch.setattr(
        MockInvalidBleakClient,
        "_response",
        lambda _s, _c, _d: bytearray(b"\x55\xaa\xeb\x90\x05")
        + bytearray(295),  # invalid frame type (0x5)
    )

    patch_bleak_client(MockInvalidBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()
    assert not result

    await bms.disconnect()


async def test_oversized_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, protocol_type
) -> None:
    """Test data update with BMS returning oversized data, result shall still be ok."""

    monkeypatch.setattr(MockOversizedBleakClient, "_FRAME", _PROTO_DEFS[protocol_type])

    patch_bleak_client(MockOversizedBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]

    await bms.disconnect()


async def test_invalid_device(patch_bleak_client) -> None:
    """Test data update with BMS returning invalid data."""

    patch_bleak_client(MockWrongBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}

    with pytest.raises(
        ConnectionError, match=r"^Failed to detect characteristics from.*"
    ):
        result = await bms.async_update()

    assert not result

    await bms.disconnect()


async def test_non_stale_data(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test if BMS class is reset if connection is reset."""

    patch_bms_timeout()

    monkeypatch.setattr(MockJikongBleakClient, "_FRAME", _PROTO_DEFS["JK02_32S"])

    orig_response = MockJikongBleakClient._response
    monkeypatch.setattr(
        MockJikongBleakClient,
        "_response",
        lambda _s, _c, _d: bytearray(b"\x55\xaa\xeb\x90\x05")
        + bytearray(10),  # invalid frame type (0x5)
    )

    patch_bleak_client(MockJikongBleakClient)

    bms = BMS(generate_ble_device())

    # run an update which provides half a valid message and then disconnects
    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()
    assert not result
    await bms.disconnect()

    # restore working BMS responses and run a test again to see if stale data is kept
    monkeypatch.setattr(MockJikongBleakClient, "_response", orig_response)

    assert await bms.async_update() == _RESULT_DEFS["JK02_32S"]


@pytest.fixture(
    name="problem_response",
    params=[
        (bytearray(b"\x01\x00"), "first_bit"),
        (bytearray(b"\x00\x80"), "last_bit"),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[bytearray, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    protocol_type: str,
    problem_response: tuple[bytearray, str],
) -> None:
    """Test data update with BMS returning system problem flags."""

    def frame_update(data: bytearray, update: bytearray, pos: int) -> None:
        data[pos : pos + 2] = update
        data[-1] = (int(data[-1]) + sum(update)) & 0xFF

    protocol_def: dict[str, dict[str, bytearray]] = deepcopy(_PROTO_DEFS)
    # set error flags in the copy

    frame_update(
        protocol_def[protocol_type]["cell"],
        problem_response[0],
        136 if protocol_type == "JK02_24S" else 166,
    )

    monkeypatch.setattr(MockJikongBleakClient, "_FRAME", protocol_def[protocol_type])

    patch_bleak_client(MockJikongBleakClient)

    bms = BMS(generate_ble_device(), False)

    assert await bms.async_update() == _RESULT_DEFS[protocol_type] | {
        "problem": True,
        "problem_code": 1 << (0 if problem_response[1] == "first_bit" else 15),
    }

    await bms.disconnect()



"""Example of using the aiobmsble library to find a BLE device by name and print its sensor data.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
import logging
from typing import Final

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from aiobmsble import BMSSample
from aiobmsble.bms.dummy_bms import BMS  # TODO: use the right BMS class for your device

NAME: Final[str] = "BT Device Name"  # TODO: replace with the name of your BLE device

# Configure logging
logging.basicConfig(level=logging.INFO)
logger: logging.Logger = logging.getLogger(__name__)


async def main(dev_name: str) -> None:
    """Find a BLE device by name and update its sensor data."""

    device: BLEDevice | None = await BleakScanner.find_device_by_name(dev_name)
    if device is None:
        logger.error("Device '%s' not found.", dev_name)
        return

    logger.info("Found device: %s (%s)", device.name, device.address)
    try:
        async with BMS(ble_device=device) as bms:
            logger.info("Updating BMS data...")
            data: BMSSample = await bms.async_update()
            logger.info("BMS data: %s", repr(data).replace(", ", ",\n\t"))
    except BleakError as ex:
        logger.error("Failed to update BMS: %s", type(ex).__name__)


if __name__ == "__main__":
    asyncio.run(main(NAME))  # pragma: no cover



"""Test the package main script."""
import argparse
import asyncio
from collections.abc import Callable
from logging import DEBUG, INFO
import sys
from typing import Any, Final
from unittest import mock

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.exc import BleakError
import pytest

from aiobmsble import BMSSample
import aiobmsble.__main__ as main_mod
from aiobmsble.bms.dummy_bms import BMS as DummyBMS
from aiobmsble.test_data import adv_dict_to_advdata


async def mock_discover(
    timeout: float = 5.0, *, return_adv: bool = False, **kwargs: Any
) -> dict[str, tuple[BLEDevice, AdvertisementData]]:
    """Mock BleakScanner to avoid actual BLE scanning."""
    mock_mac_unknown: Final[str] = "00:00:00:00:00:00"
    mock_mac: Final[str] = "11:22:33:44:55:66"
    mock_device: BLEDevice = BLEDevice(mock_mac, "Dummy BMS", None)
    mock_adv: AdvertisementData = adv_dict_to_advdata({"local_name": "dummy"})
    assert timeout >= 0, "timeout cannot be negative."
    assert return_adv, "mock only works with advertisement info."
    return {
        mock_mac_unknown: (
            BLEDevice(mock_mac_unknown, "Unknown Device", None),
            adv_dict_to_advdata({"local_name": "unknown_device"}),
        ),
        mock_mac: (mock_device, mock_adv),
    }


@pytest.fixture(name="mock_setup_logging")
def setup_logging():
    """Unittest mock for setup_logging to check calls to it."""
    with mock.patch.object(main_mod, "setup_logging") as m:
        yield m


@pytest.fixture(name="mock_asyncio_run")
def asyncio_run():
    """Unittest mock for asyncio_run to check calls to it."""
    with mock.patch("asyncio.run") as m:
        m.side_effect = lambda coro: asyncio.new_event_loop().run_until_complete(coro)
        yield m


async def test_detect_bms(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Verify log output for working BMS update query."""

    monkeypatch.setattr("aiobmsble.__main__.BleakScanner.discover", mock_discover)
    patch_bleak_client()
    with caplog.at_level(INFO):
        await main_mod.detect_bms()
    assert "Found matching BMS type: Dummy Manufacturer dummy model" in caplog.text
    assert (
        "BMS data: {'voltage': 12,\n\t'current': 1.5,\n\t'temperature': 27.182,\n"
        "\t'power': 18.0,\n\t'battery_charging': True,\n\t'problem': False}\n"
        in caplog.text
    )


async def test_scan_devices_fail(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    """Verify log output for working BMS update query."""

    async def mock_discover_fail(
        timeout: float = 5.0, *, return_adv: bool = False, **kwargs: Any
    ) -> dict[str, tuple[BLEDevice, AdvertisementData]]:
        raise BleakError("No BT adapters.")

    monkeypatch.setattr("aiobmsble.__main__.BleakScanner.discover", mock_discover_fail)

    with caplog.at_level(INFO):
        await main_mod.scan_devices()
    assert "Could not scan for BT devices: No BT adapters." in caplog.text


async def test_bms_fail(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check that an error message is given if BMS update query fails (TimeoutError)."""

    async def mock_async_update(self) -> BMSSample:
        raise TimeoutError

    monkeypatch.setattr("aiobmsble.__main__.BleakScanner.discover", mock_discover)
    monkeypatch.setattr("aiobmsble.bms.dummy_bms.BMS._async_update", mock_async_update)
    patch_bleak_client()
    with caplog.at_level(INFO):
        await main_mod.detect_bms()
    assert "Found matching BMS type: Dummy Manufacturer dummy model" in caplog.text
    assert "Failed to query BMS: TimeoutError" in caplog.text


async def test_bms_retry_with_secret(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Verify that a failed query will be retried when a BMS accepts a secret.

    The first call to ``_async_update`` simulates a timeout while no secret is
    supplied. ``detect_bms`` should catch the error, prompt for a secret via
    :func:`getpass.getpass`, and attempt the query again with the provided
    value.  The second attempt returns a valid sample which is logged.
    """

    # prepare BLE scanning and client patching as in other tests
    monkeypatch.setattr("aiobmsble.__main__.BleakScanner.discover", mock_discover)

    DummyBMS.accept_secret = True
    orig_init = DummyBMS.__init__

    def init_with_secret(self, ble_device, keep_alive=True, secret="") -> None:
        # call original initializer and remember the secret
        orig_init(self, ble_device, keep_alive)
        self._secret = secret

    monkeypatch.setattr(DummyBMS, "__init__", init_with_secret)

    attempts: list[str] = []

    async def fake_update(self) -> BMSSample:
        # record the secret used for each invocation
        attempts.append(getattr(self, "_secret", ""))
        if not getattr(self, "_secret", ""):
            raise TimeoutError
        return {"voltage": 99}

    monkeypatch.setattr(DummyBMS, "_async_update", fake_update)

    # patch getpass to avoid interactive prompt
    monkeypatch.setattr("getpass.getpass", lambda prompt="": "secret123")

    patch_bleak_client()
    with caplog.at_level(INFO):
        await main_mod.detect_bms()

    assert "Failed to query BMS: TimeoutError" in caplog.text
    assert "Querying BMS with secret..." in caplog.text
    assert "BMS data" in caplog.text
    # ensure we attempted without and then with the secret
    assert attempts == ["", "secret123"]


def test_main_parses_logfile_and_verbose(
    monkeypatch: pytest.MonkeyPatch,
    mock_setup_logging: mock.MagicMock | mock.AsyncMock,
    mock_asyncio_run: mock.MagicMock | mock.AsyncMock,
) -> None:
    """Check that command line parses log file option and verbosity level."""

    async def patch_scan_devices() -> dict[str, tuple[BLEDevice, AdvertisementData]]:
        return {}

    monkeypatch.setattr(sys, "argv", ["prog", "-l", "test.log", "-v"])
    monkeypatch.setattr(main_mod, "scan_devices", patch_scan_devices)
    main_mod.main()
    args = mock_setup_logging.call_args[0][0]
    assert mock_setup_logging.called
    assert mock_asyncio_run.called
    assert isinstance(args, argparse.Namespace)
    assert args.logfile == "test.log"
    assert args.verbose


@mock.patch("aiobmsble.__main__.logger")
def test_logging_wo_file(mock_logger) -> None:
    """Check default log level."""
    args = argparse.Namespace(verbose=False, logfile=None)

    main_mod.setup_logging(args)

    assert mock_logger.setLevel.call_args[0][0] == INFO
    mock_logger.addHandler.assert_not_called()


@mock.patch("aiobmsble.__main__.logger")
def test_verbose_logging_wo_file(mock_logger) -> None:
    """Check verbose log level."""
    args = argparse.Namespace(verbose=True, logfile=None)

    main_mod.setup_logging(args)

    assert mock_logger.setLevel.call_args[0][0] == DEBUG
    mock_logger.addHandler.assert_not_called()


@mock.patch("aiobmsble.__main__.logging.FileHandler")
@mock.patch("aiobmsble.__main__.logger")
def test_logging_with_logfile(mock_logger, mock_file_handler_cls) -> None:
    """Check that logging goes to file if log file is given."""
    mock_file_handler = mock.MagicMock()
    mock_file_handler_cls.return_value = mock_file_handler

    args = argparse.Namespace(verbose=False, logfile="test.log")

    main_mod.setup_logging(args)

    mock_file_handler.setLevel.assert_called_once_with(INFO)
    mock_file_handler.setFormatter.assert_called_once()
    mock_logger.addHandler.assert_called_once_with(mock_file_handler)
    mock_logger.setLevel.assert_called_once_with(INFO)



"""Test the Neey BMS implementation."""

import asyncio
from collections.abc import Buffer
from copy import deepcopy
from typing import Final, cast
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.basebms import crc_sum
from aiobmsble.bms.neey_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 29

_PROTO_DEFS: Final[dict[str, bytearray]] = {
    "dev": bytearray(
        b"\x55\xaa\x11\x01\x01\x00\x64\x00\x47\x57\x2d\x32\x34\x53\x34\x45\x42\x00\x00\x00"
        b"\x00\x00\x00\x00\x48\x57\x2d\x32\x2e\x38\x2e\x30\x5a\x48\x2d\x31\x2e\x32\x2e\x33"
        b"\x56\x31\x2e\x30\x2e\x30\x00\x00\x32\x30\x32\x32\x30\x35\x33\x31\x05\x00\x00\x00"
        b"\x01\x91\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\xff"
    ),
    "ack": bytearray(
        # b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x41"
        # b"\x54\x0d\x0a"
    ),  # ACKnowledge message with attached AT\r\n message (needs to be filtered)
    "cell": bytearray(
        b"\x55\xaa\x11\x01\x02\x00\x2c\x01\x38\xe7\xfa\x50\x40\xb6\x04\x51\x40\x85\x0e\x51"
        b"\x40\xf0\x05\x51\x40\xb6\x04\x51\x40\x75\x1e\x51\x40\x7f\x4f\x51\x40\x43\x02\x51"
        b"\x40\x1c\x3d\x51\x40\x78\x6a\x51\x40\xfe\x82\x51\x40\x16\x7e\x51\x40\xbc\x76\x51"
        b"\x40\x16\x7e\x51\x40\x8b\x80\x51\x40\xca\x66\x51\x40\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x35\x93\x24\x3e\x68\x94\x26\x3e\x3d\x25\x1b\x3e\x90\x8e\x1b"
        b"\x3e\xb3\xf3\x23\x3e\x2e\x91\x25\x3e\xc6\x1b\x1a\x3e\x4a\x7c\x1c\x3e\x6f\x1b\x1a"
        b"\x3e\xc2\x43\x1b\x3e\x85\x1e\x18\x3e\x4b\x27\x19\x3e\x5e\xdf\x18\x3e\xd0\xeb\x1a"
        b"\x3e\xe6\xd4\x18\x3e\x0c\xfe\x18\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\xde\x40\x51\x42\xde\x40\x51\x40\x00\x17\x08\x3c\x0a\x00\x0f\x05\x19\xa1\x82"
        b"\xc0\xc3\xf5\x48\x42\xc3\xf5\x48\x42\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x2e\x09\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb6\xff"
    ),
}

_RESULT_DEFS: Final[BMSSample] = {
    "delta_voltage": 0.008,
    "temperature": 50.24,
    "voltage": 52.313,
    "balance_current": -4.082,
    "cell_count": 16,
    "cell_voltages": [
        3.265,
        3.266,
        3.267,
        3.266,
        3.266,
        3.267,
        3.27,
        3.266,
        3.269,
        3.272,
        3.274,
        3.273,
        3.273,
        3.273,
        3.273,
        3.272,
    ],
    "balancer": True,
    "temp_values": [50.24, 50.24],
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockNeeyBleakClient(MockBleakClient):
    """Emulate a Neey BMS BleakClient."""

    HEAD_CMD: Final = bytearray(b"\xaa\x55\x11\x01")
    DEV_INFO: Final = bytearray(b"\x01")
    CELL_INFO: Final = bytearray(b"\x02")
    TAIL: Final = 0xFF
    _FRAME: dict[str, bytearray] = {}

    _task: asyncio.Task[None] | None = None

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        frame: Final[bytearray] = bytearray(data)
        if (
            char_specifier != "ffe1"
            or frame[19] != self.TAIL
            or not frame.startswith(self.HEAD_CMD)
        ):
            return bytearray()
        if frame[4:5] == self.CELL_INFO:
            return self._FRAME["cell"]
        if frame[4:5] == self.DEV_INFO:
            return self._FRAME["dev"]

        return bytearray()

    async def _send_confirm(self) -> None:
        assert self._notify_callback, "send confirm called but notification not enabled"
        await asyncio.sleep(0)
        self._notify_callback(
            "MockNeeyBleakClient",
            b"\xaa\x55\x90\xeb\xc8\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44",
        )

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"

        resp: Final[bytearray] = self._response(char_specifier, data)
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockNeeyBleakClient", notify_data)


class MockStreamBleakClient(MockNeeyBleakClient):
    """Mock Neey BMS that already sends battery data (no request required)."""

    async def _send_all(self) -> None:
        assert (
            self._notify_callback
        ), "send_all frames called but notification not enabled"
        for resp in self._FRAME.values():
            self._notify_callback("MockNeeyBleakClient", resp)
            await asyncio.sleep(0)

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""

        assert (
            self._notify_callback
        ), "write to characteristics but notification not enabled"
        if bytearray(data).startswith(
            self.HEAD_CMD + self.DEV_INFO
        ):  # send all responses as a series
            self._task = asyncio.create_task(self._send_all())

    async def disconnect(self) -> None:
        """Mock disconnect and wait for send task."""
        if self._task and not self._task.done():
            await asyncio.wait_for(self._task, 0.1)
            assert self._task.done(), "send task still running!"
        await super().disconnect()


class MockOversizedBleakClient(MockNeeyBleakClient):
    """Emulate a Neey BMS BleakClient returning wrong data length."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        return super()._response(char_specifier, data) + bytearray(6)


async def test_update(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, keep_alive_fixture: bool
) -> None:
    """Test Neey BMS data update."""

    monkeypatch.setattr(MockNeeyBleakClient, "_FRAME", _PROTO_DEFS)
    patch_bleak_client(MockNeeyBleakClient)
    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    assert await bms.async_update() == _RESULT_DEFS
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(monkeypatch, patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    monkeypatch.setattr(MockNeeyBleakClient, "_FRAME", _PROTO_DEFS)
    patch_bleak_client(MockNeeyBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "model": "GW-24S4EB",
        "sw_version": "ZH-1.2.3",
        "hw_version": "HW-2.8.0",
    }


async def test_stream_update(monkeypatch, patch_bleak_client, caplog) -> None:
    """Test Neey BMS data update."""

    _frames: dict[str, bytearray] = deepcopy(_PROTO_DEFS)
    monkeypatch.setattr(MockStreamBleakClient, "_FRAME", _frames)
    patch_bleak_client(MockStreamBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS
    assert bms._msg_event.is_set() is False
    assert "requesting cell info" in caplog.text

    _client: MockStreamBleakClient = cast(MockStreamBleakClient, bms._client)
    _cell_frame: bytearray = _frames["cell"]
    _cell_frame[216] = 0x8  # modify data to ensure new read
    _cell_frame[-2] = crc_sum(_cell_frame[:-2])
    caplog.clear()
    await _client._send_all()

    # query again to see if updated streaming data is used
    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": 0x8,
        "balancer": False,
    }
    assert bms._msg_event.is_set() is False, "BMS does not request fresh data"
    assert "requesting cell info" not in caplog.text, "BMS did not use streaming data"


@pytest.fixture(
    name="wrong_response",
    params=[
        (_PROTO_DEFS["dev"][:-2] + b"\x00\xff", "wrong_CRC"),
        (bytearray(b"\x55\xaa\xeb\x90\x05") + bytes(295), "wrong_frame_type"),
        (_PROTO_DEFS["dev"][:-1] + b"\x00", "wrong_EOF"),
    ],
    ids=lambda param: param[1],
)
def faulty_response(request: pytest.FixtureRequest) -> bytearray:
    """Return faulty response frame."""
    assert isinstance(request.param[0], bytearray)
    return request.param[0]


async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytearray,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockNeeyBleakClient, "_response", lambda _s, _c, _d: wrong_response
    )
    patch_bleak_client(MockNeeyBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_oversized_response(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client
) -> None:
    """Test data update with BMS returning oversized data, result shall still be ok."""

    monkeypatch.setattr(MockOversizedBleakClient, "_FRAME", _PROTO_DEFS)

    patch_bleak_client(MockOversizedBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS

    await bms.disconnect()


async def test_non_stale_data(
    monkeypatch, patch_bleak_client, patch_bms_timeout
) -> None:
    """Test if BMS class is reset if connection is reset."""

    patch_bms_timeout()

    monkeypatch.setattr(MockNeeyBleakClient, "_FRAME", _PROTO_DEFS)

    orig_response = MockNeeyBleakClient._response
    monkeypatch.setattr(
        MockNeeyBleakClient,
        "_response",
        lambda _s, _c, _d: bytearray(b"\x55\xaa\xeb\x90\x05") + bytearray(10),
    )  # invalid frame type (0x5)

    patch_bleak_client(MockNeeyBleakClient)

    bms = BMS(generate_ble_device())

    # run an update which provides half a valid message and then disconnects
    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()
    assert not result
    await bms.disconnect()

    # restore working BMS responses and run a test again to see if stale data is kept
    monkeypatch.setattr(MockNeeyBleakClient, "_response", orig_response)

    assert await bms.async_update() == _RESULT_DEFS


@pytest.fixture(
    name="problem_response",
    params=[
        (0x01, "Wrong cell count"),
        # (0x02, "AcqLine Res test"),
        (0x03, "AcqLine Res exceed"),
        # (0x04, "Systest Completed"),
        # (0x05, "Balancing"),
        # (0x06, "Balancing finished"),
        (0x07, "Low voltage"),
        (0x08, "System Overtemp"),
        (0x09, "Host fails"),
        (0x0A, "Low battery voltage - balancing stopped"),
        (0x0B, "Temperature too high - balancing stopped"),
        # (0x0C, "Self-test completed"),
    ],
    ids=lambda param: param[1],
)
def prb_response(request: pytest.FixtureRequest) -> tuple[int, str]:
    """Return faulty response frame."""
    assert isinstance(request.param, tuple)
    return request.param


async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: tuple[int, str],
) -> None:
    """Test data update with BMS returning system problem flags."""

    def frame_update(data: bytearray, update: int) -> None:
        data[-2] = (data[-2] + update - data[216]) & 0xFF
        data[216] = update

    protocol_def: dict[str, bytearray] = deepcopy(_PROTO_DEFS)
    # set error flags in the copy

    frame_update(protocol_def["cell"], problem_response[0])

    monkeypatch.setattr(MockNeeyBleakClient, "_FRAME", protocol_def)

    patch_bleak_client(MockNeeyBleakClient)

    bms = BMS(generate_ble_device(), False)

    assert await bms.async_update() == _RESULT_DEFS | {
        "problem": True,
        "problem_code": problem_response[0],
        "balancer": False,
    }

    await bms.disconnect()



"""Test the LiPower BMS implementation."""

from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.lipower_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 32

_PROTO_DEFS: Final[dict[int, bytearray]] = {
    0x22: bytearray(
        b"\x22\x03\x10\x00\x76\x00\x63\x05\xcf\x00\x16\x00\x01\x00\x08\x00\x89\x00\x01\x9e\xcf"
    ),  # 13.7V, 99%, 5354520s, -0.08A, -1W
    0x0B: bytearray(
        b"\x0b\x03\x12\x00\x70\x00\x5e\x00\x35\x00\x2d\x00\x01\x00\xc9\x00\x85\x00\x1a\x00\x02\x47\x16"
    ),  # 13.3V, 94%, 193500s, -2.01A, -26W, protocol version 0x02
}

_RESULT_DEFS: Final[dict[int, BMSSample]] = {
    0x22: {
        "voltage": 13.7,
        "current": -0.08,
        "battery_level": 99,
        "cycle_charge": 118,
        "cycle_capacity": 1616.6,
        "power": -1.096,
        "runtime": 5354520,
        "battery_charging": False,
        "problem": False,
    },
    0x0B: {
        "voltage": 13.3,
        "current": -2.01,
        "battery_level": 94,
        "cycle_charge": 112,
        "cycle_capacity": 1489.6,
        "power": -26.733,
        "runtime": 193500,
        "battery_charging": False,
        "problem": False,
    },
}


@pytest.fixture(
    name="protocol_type",
    params=[0x22, 0x0B],
)
def proto(request: pytest.FixtureRequest) -> int:
    """Protocol fixture."""
    assert isinstance(request.param, int)
    return request.param


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockLiPwrBleakClient(MockBleakClient):
    """Emulate a LiPower BMS BleakClient."""

    _RESP: Final[bytearray] = _PROTO_DEFS[0x22]

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if not isinstance(char_specifier, str) or char_specifier != "ffe1":
            return bytearray()
        addr: int = int.from_bytes(bytes(data)[3:5])
        if addr == 0x0 and bytes(data)[0:2] == MockLiPwrBleakClient._RESP[0:2]:
            return MockLiPwrBleakClient._RESP
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None

        for notify_data in [
            self._response(char_specifier, data)[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._response(char_specifier, data)), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockLiPwrBleakClient", notify_data)


async def test_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    protocol_type: int,
    keep_alive_fixture: bool,
) -> None:
    """Test LiPower BMS data update."""

    patch_bms_timeout()
    monkeypatch.setattr(MockLiPwrBleakClient, "_RESP", _PROTO_DEFS[protocol_type])
    patch_bleak_client(MockLiPwrBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS[protocol_type]

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"\x10\x03\x10\x00\x76\x00\x63\x05\xcf\x00\x16\x00\x01\x00\x08\x00\x89\x00\x01\xa8\x73",
        b"\x22\x03\x10\x00\x76\x00\x63\x05\xcf\x00\x16\x00\x01\x00\x08\x00\x89\x00\x01\x00\xcf",
        b"\x22\x03\x11\x00\x76\x00\x63\x05\xcf\x00\x16\x00\x01\x00\x08\x00\x89\x00\x01\xcf\x5f",
        b"",
    ],
    ids=["wrong_SOF", "wrong_CRC", "wrong_len", "empty"],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockLiPwrBleakClient, "_response", lambda x, y, z: bytearray(wrong_response)
    )
    patch_bleak_client(MockLiPwrBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()



"""Test the Wattstunde Nova BMS implementation."""

from collections.abc import Buffer
from typing import Final, cast
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.ws_nova_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 20

_PROTO_DEFS: Final[bytearray] = bytearray(
    b"\x3a\x32\x31\x37\x34\x32\x31\x32\x30\x43\x43\x32\x30\x32\x31\x32\x32\x34\x39\x33"
    b"\x38\x38\x30\x43\x31\x32\x43\x39\x35\x32\x43\x39\x35\x32\x43\x39\x39\x32\x43\x39"
    b"\x34\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32"
    b"\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32"
    b"\x30\x32\x30\x32\x30\x32\x30\x32\x30\x44\x30\x32\x32\x32\x30\x32\x30\x31\x31\x31"
    b"\x31\x31\x31\x31\x31\x33\x30\x45\x38\x41\x30\x32\x30\x32\x35\x44\x30\x35\x46\x44"
    b"\x46\x44\x46\x44\x46\x31\x32\x46\x37\x32\x30\x32\x35\x32\x34\x32\x30\x32\x35\x31"
    b"\x34\x32\x30\x32\x33\x32\x44\x36\x30\x32\x30\x32\x31\x39\x31\x32\x31\x32\x30\x32"
    b"\x33\x32\x44\x36\x30\x32\x30\x32\x31\x32\x35\x44\x30\x32\x32\x32\x36\x37\x37\x37"
    b"\x33\x36\x45\x31\x35\x31\x35\x31\x30\x31\x30\x31\x37\x31\x32\x31\x34\x31\x32\x31"
    b"\x31\x31\x31\x31\x30\x31\x39\x31\x33\x32\x30\x32\x30\x32\x30\x32\x30\x32\x30\x32"
    b"\x30\x32\x30\x32\x30\x32\x43\x33\x34\x32\x33\x32\x37\x33\x38\x41\x41\x7e"
)

_RESULT_DEFS: Final[BMSSample] = {
    "voltage": 13.015,
    "current": -1.52,
    "battery_level": 52,
    "delta_voltage": 0.005,
    "power": -19.783,
    "runtime": 262537,
    "battery_charging": False,
    "cycle_charge": 110.849,
    "cycles": 5,
    "design_capacity": 200,
    "temp_values": [9.0, 9.0, 9.0, 9.0],
    "temperature": 9.0,
    "cell_voltages": [
        3.253,
        3.253,
        3.257,
        3.252,
    ],
    "cell_count": 4,
    "cycle_capacity": 1442.7,
    "heater": False,
    "problem": False,
    "problem_code": 0,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockWSNovaBleakClient(MockBleakClient):
    """Emulate a Wattstunde Nova BMS BleakClient."""

    _RESP: bytearray = _PROTO_DEFS

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        if char_specifier != "FFF1":
            return
        for notify_data in [
            self._RESP[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._RESP), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockStreamBleakClient", notify_data)


class MockStreamBleakClient(MockWSNovaBleakClient):
    """Emulate a Wattstunde Nova BMS BleakClient sending data in stream mode."""

    _RESP_NOTIFY: bytearray = _PROTO_DEFS

    async def _notify(self) -> None:
        assert self._notify_callback is not None
        for notify_data in [
            self._RESP_NOTIFY[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._RESP_NOTIFY), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockStreamBleakClient", notify_data)


async def test_update(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, keep_alive_fixture: bool
) -> None:
    """Test Wattstunde Nova BMS data update."""

    monkeypatch.setattr(MockWSNovaBleakClient, "_RESP", _PROTO_DEFS)
    patch_bleak_client(MockWSNovaBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_stream_update(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test Wattstunde Nova BMS stream data update."""

    patch_bleak_client(MockStreamBleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == _RESULT_DEFS
    assert bms._msg_event.is_set() is False, "BMS does not request fresh data"
    assert "requesting BMS data" in caplog.text

    _client: MockStreamBleakClient = cast(MockStreamBleakClient, bms._client)
    caplog.clear()
    _frame: bytearray = _PROTO_DEFS.copy()
    _frame[135:139] = b"2029"  # change cycles from 5 to 9
    monkeypatch.setattr(_client, "_RESP_NOTIFY", _frame)
    await _client._notify()

    # query again to see if updated streaming data is used
    assert await bms.async_update() == _RESULT_DEFS | {"cycles": 0x9}
    assert "requesting BMS data" not in caplog.text, "BMS did not use streaming data"

    caplog.clear()
    # do not automatically send data this time
    _frame = _PROTO_DEFS.copy()
    _frame[135:139] = b"2429"  # change cycles from 9 to 73
    monkeypatch.setattr(_client, "_RESP_NOTIFY", _frame)

    # query again to see if BMS recovers if no data is sent
    assert await bms.async_update() == _RESULT_DEFS
    assert "requesting BMS data" in caplog.text, "BMS did not use streaming data"


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockWSNovaBleakClient)
    bms = BMS(generate_ble_device())
    assert await bms.device_info() == {
        "serial_number": "5500724211093",
    }


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"",
        b":~",
        b"-" + _PROTO_DEFS[1:],
        _PROTO_DEFS[:-2] + b"#",
        b":" + _PROTO_DEFS[2:],
        b":z" + _PROTO_DEFS[2:],
        b":2175" + _PROTO_DEFS[5:],
    ],
    ids=[
        "empty",
        "minimal",
        "wrong_SOF",
        "wrong_EOF",
        "wrong_length",
        "wrong_encoding",
        "wrong_type",
    ],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(MockWSNovaBleakClient, "_RESP", bytearray(wrong_response))
    patch_bleak_client(MockWSNovaBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


@pytest.mark.parametrize(
    ("problem_response", "expected"),
    [
        (b"\x43\x38\x32\x30", 0x0800),
        (b"\x45\x30\x33\x30", 0x0010),
        (b"\x46\x30\x32\x34", 0x0004),
        (b"\x43\x30\x32\x38", 0x0008),
    ],
    ids=["lowT_dischrg", "overC_chrg", "high", "low"],
)
async def test_problem_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    problem_response: bytes,
    expected: int,
) -> None:
    """Test data update with BMS returning error flags."""

    _resp: bytearray = _PROTO_DEFS.copy()
    _resp[89:93] = problem_response
    monkeypatch.setattr(MockWSNovaBleakClient, "_RESP", _resp)

    patch_bleak_client(MockWSNovaBleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = await bms.async_update()
    assert result == _RESULT_DEFS | {"problem": True, "problem_code": expected}

    await bms.disconnect()



"""Test the Dummy BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic

from aiobmsble.bms.dummy_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockDummyBleakClient(MockBleakClient):
    """Emulate a Dummy BMS BleakClient."""

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None

        self._notify_callback(
            "MockDummyBleakClient", bytearray(f"response to '{data!s}'", "UTF-8")
        )


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test Dummy BMS data update."""

    patch_bleak_client(MockDummyBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == {
        "voltage": 12,
        "current": 1.5,
        "temperature": 27.182,
        "battery_charging": True,
        "power": 18,
        "problem": False,
    }

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    patch_bleak_client(MockDummyBleakClient)
    bms = BMS(generate_ble_device())
    assert {"default_manufacturer", "default_model"}.issubset(await bms.device_info())



"""Test the AG Automotive (E&J) BMS implementation."""

from collections.abc import Buffer
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str

from aiobmsble.bms.ag_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests

BT_FRAME_SIZE = 20


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockAGBleakClient(MockBleakClient):
    """Emulate a AG Automotive (E&J) BMS BleakClient."""

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("00008001-0000-1000-8000-57616c6b697a"):
            return bytearray()
        cmd: int = int(bytearray(data)[3:5], 16)
        if cmd == 0x02:
            return bytearray(
                b"\x3a\x30\x30\x38\x32\x35\x30\x30\x30\x38\x30\x30\x30\x30\x31\x30\x31\x43\x30\x30"
                b"\x30\x30\x30\x38\x38\x30\x43\x45\x30\x30\x43\x45\x31\x30\x43\x44\x46\x30\x43\x45"
                b"\x31\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30"
                b"\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x30\x31\x34\x45\x32\x45\x32"
                b"\x41\x32\x41\x32\x41\x46\x30\x30\x32\x30\x30\x30\x30\x30\x30\x30\x30\x33\x46\x30"
                b"\x30\x33\x46\x33\x43\x30\x32\x7e"
            )  # 63 cycles, -3.3A, 60% battery, 4 cells at 3.29V, delta 0.002, temp 6°C
        if cmd == 0x10:
            return bytearray(
                b"\x3a\x30\x30\x39\x30\x33\x31\x30\x30\x31\x45\x30\x30\x30\x30\x30\x32\x34\x45\x30"
                b"\x33\x45\x38\x30\x33\x45\x38\x41\x31\x7e"
            )
        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        for notify_data in [
            self._response(char_specifier, data)[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._response(char_specifier, data)), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockAGBleakClient", notify_data)


async def test_update(patch_bleak_client, keep_alive_fixture: bool) -> None:
    """Test AG Automotive (E&J) BMS data update."""

    patch_bleak_client(MockAGBleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == {
        "voltage": 13.185,
        "current": -3.34,
        "battery_level": 60,
        "battery_charging": False,
        "chrg_mosfet": True,
        "cycle_capacity": 777.915,
        "cycle_charge": 59.0,
        "cycles": 63,
        "cell_count": 4,
        "cell_voltages": [3.296, 3.297, 3.295, 3.297],
        "delta_voltage": 0.002,
        "dischrg_mosfet": True,
        "power": -44.038,
        "temperature": 6.0,
        "temp_values": [6.0],
        "problem": False,
        "problem_code": 0,
        "balancer": 0,
        "heater": False,
        "runtime": 63592,
    }

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()



"""Test the Super-B v2 BMS implementation."""

from collections.abc import Awaitable, Buffer, Callable
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSSample
from aiobmsble.bms.superb_v2_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient
from tests.test_basebms import BMSBasicTests, verify_device_info

BT_FRAME_SIZE = 32

RESP: dict[bytes, bytearray] = {
    b"\x21\x54\x00": bytearray(
        b"\x00\x23\x04\x13\xff\xff\x81\xe3\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12\x64\x00\x01\x2f\x29"
    )
}


_RESULT_DEFS: Final[BMSSample] = {
    # "balancer": False,
    "battery_charging": False,
    "battery_health": 100,
    "battery_level": 35,
    "cycles": 18,
    "current": -2.465,
    "power": -32.555,
    "problem": False,
    "runtime": 77609,
    "voltage": 13.207,
}


class TestBasicBMS(BMSBasicTests):
    """Test the basic BMS functionality."""

    bms_class = BMS


class MockSuperBv2BleakClient(MockBleakClient):
    """Emulate a Super-B v2 BMS BleakClient."""

    _QUERY_RESP: dict[bytes, bytearray] = RESP
    _NOTIFY_RESP: bytearray = bytearray(
        b"\x02\x0d\x00\x00\x00\x00\xff\xff\xf6\x5f\x33\x97\x00\x00\x09\x64\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
    )

    def _send_info(self) -> None:
        assert self._notify_callback is not None
        for notify_data in [
            self._NOTIFY_RESP[i : i + BT_FRAME_SIZE]
            for i in range(0, len(self._NOTIFY_RESP), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockSuperBv2BleakClient", notify_data)

    def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, cmd: bytes
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) != normalize_uuid_str("cf9ccdfa-eee9-43ce-87a5-82b54af5324e"):
            return bytearray()

        return self._QUERY_RESP.get(cmd, bytearray())

    @property
    def is_connected(self) -> bool:
        """Mock connected."""
        if self._connected:
            self._send_info()  # patch to provide data when not reconnecting
        return self._connected

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
        ],
        **kwargs,
    ) -> None:
        """Mock start_notify."""
        await super().start_notify(char_specifier, callback)
        self._send_info()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None

        resp: bytearray = self._response(char_specifier, bytes(data))
        for notify_data in [
            resp[i : i + BT_FRAME_SIZE] for i in range(0, len(resp), BT_FRAME_SIZE)
        ]:
            self._notify_callback("MockSuperBv2BleakClient", notify_data)


async def test_update(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client, keep_alive_fixture: bool
) -> None:
    """Test Super-B v2 BMS data update."""

    patch_bleak_client(MockSuperBv2BleakClient)

    bms = BMS(generate_ble_device(), keep_alive_fixture)

    assert await bms.async_update() == _RESULT_DEFS

    # query again to check already connected state
    await bms.async_update()
    assert bms.is_connected is keep_alive_fixture

    await bms.disconnect()


async def test_update_chrg(monkeypatch: pytest.MonkeyPatch, patch_bleak_client) -> None:
    """Test Super-B v2 BMS data update while charging."""

    monkeypatch.setattr(
        MockSuperBv2BleakClient,
        "_QUERY_RESP",
        {
            b"\x21\x54\x00": bytearray(
                b"\x00\x23\x04\x13\x00\x01\xdf\xd9\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12\x64\xff\xff\xff\xff"
            )
        },
    )
    monkeypatch.setattr(
        MockSuperBv2BleakClient,
        "_NOTIFY_RESP",
        bytearray(
            b"\x02\x0e\x00\x00\x00\x00\x00\x00\x24\x12\x33\xd9\x00\x00\x09\x64\x00\x00\x00\x00\x00\x00\x00\x00"
        ),
    )
    patch_bleak_client(MockSuperBv2BleakClient)

    bms = BMS(generate_ble_device())

    assert await bms.async_update() == {
        # "balancer": False,
        "battery_charging": True,
        "battery_health": 100,
        "battery_level": 35,
        "cycles": 18,
        "current": 9.234,
        "power": 122.563,
        "problem": False,
        "voltage": 13.273,
    }

    await bms.disconnect()


async def test_device_info(patch_bleak_client) -> None:
    """Test that the BMS returns initialized dynamic device information."""
    await verify_device_info(patch_bleak_client, MockSuperBv2BleakClient, BMS)


@pytest.mark.parametrize(
    ("wrong_response"),
    [
        b"",
        MockSuperBv2BleakClient._NOTIFY_RESP[:-1],
        MockSuperBv2BleakClient._NOTIFY_RESP + b"\x00",
    ],
    ids=["empty", "too_short", "too_long"],
)
async def test_invalid_response(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    patch_bms_timeout,
    wrong_response: bytes,
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout()
    monkeypatch.setattr(
        MockSuperBv2BleakClient, "_NOTIFY_RESP", bytearray(wrong_response)
    )
    patch_bleak_client(MockSuperBv2BleakClient)

    bms = BMS(generate_ble_device())

    result: BMSSample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()
